#ifndef ZASM_ALLOC_H
#define ZASM_ALLOC_H 1

#define ALLOC_DEF_COUNT 64

// pure virtual base class
#define ALLOC_T(O) ((struct alloc_t *)(O))
struct alloc_t
{
  // public
  int i;
  struct alloc_t *previous;  
};

void alloc_fill(struct alloc_t *alloc);
void alloc_delete(struct alloc_t *alloc);

#define ALLOC_DECLARE(TYPE, COUNT) \
struct TYPE ## _alloc_t\
{\
  struct alloc_t alloc;\
  struct TYPE ## _t data[COUNT];\
};\
\
struct TYPE ## _alloc_t *TYPE ## _alloc_new();\
void TYPE ## _alloc_fill(struct TYPE ## _alloc_t *alloc);\
struct TYPE ## _t *TYPE ## _alloc_get(struct TYPE ## _alloc_t **alloc);\
void TYPE ## _alloc_reset(struct TYPE ## _alloc_t *alloc);\
void TYPE ## _alloc_delete(struct TYPE ## _alloc_t *alloc);\
void TYPE ## _alloc_free(struct TYPE ## _alloc_t *alloc);

#include <zasm/memory.h>

#define ALLOC_DEFINE(TYPE, COUNT) \
struct TYPE ## _alloc_t *TYPE ## _alloc_new()\
{\
  struct TYPE ## _alloc_t *val;\
\
  ALLOCATE(struct TYPE ## _alloc_t, val, 1);\
  TYPE ## _alloc_fill(val);\
  return val;\
}\
\
void TYPE ## _alloc_fill(struct TYPE ## _alloc_t *alloc)\
{\
  alloc_fill(&(alloc->alloc));\
}\
\
struct TYPE ## _t *TYPE ## _alloc_get(struct TYPE ## _alloc_t **alloc)\
{\
  struct TYPE ## _alloc_t *temp;\
  if ((!(*alloc)) || ((*alloc)->alloc.i >= COUNT))\
    {\
      temp = TYPE ## _alloc_new();\
      temp->alloc.previous = &((*alloc)->alloc);\
      *alloc = temp;\
    }\
  (*alloc)->alloc.i++;\
  return &((*alloc)->data[(*alloc)->alloc.i - 1]);\
}\
\
void TYPE ## _alloc_delete(struct TYPE ## _alloc_t *alloc)\
{\
  if (alloc->alloc.previous)\
  {\
    TYPE ## _alloc_free((struct TYPE ## _alloc_t *)(alloc->alloc.previous));\
    alloc->alloc.previous = NULL;\
  }\
}\
\
void TYPE ## _alloc_reset(struct TYPE ## _alloc_t *alloc)\
{\
  TYPE ## _alloc_delete(alloc);\
  TYPE ## _alloc_fill(alloc);\
}\
\
void TYPE ## _alloc_free(struct TYPE ## _alloc_t *alloc)\
{\
  if (!alloc)\
    return;\
  TYPE ## _alloc_delete(alloc);\
  free(alloc);\
}

#define ALLOC_DEFINE_DELETE(TYPE, COUNT) \
struct TYPE ## _alloc_t *TYPE ## _alloc_new()\
{\
  struct TYPE ## _alloc_t *val;\
\
  ALLOCATE(struct TYPE ## _alloc_t, val, 1);\
  TYPE ## _alloc_fill(val);\
  return val;\
}\
\
void TYPE ## _alloc_fill(struct TYPE ## _alloc_t *alloc)\
{\
  alloc_fill(&(alloc->alloc));\
}\
\
struct TYPE ## _t *TYPE ## _alloc_get(struct TYPE ## _alloc_t **alloc)\
{\
  struct TYPE ## _alloc_t *temp;\
  if ((!(*alloc)) || ((*alloc)->alloc.i >= COUNT))\
    {\
      temp = TYPE ## _alloc_new();\
      temp->alloc.previous = &((*alloc)->alloc);\
      *alloc = temp;\
    }\
  (*alloc)->alloc.i++;\
  return &((*alloc)->data[(*alloc)->alloc.i - 1]);\
}\
\
void TYPE ## _alloc_reset(struct TYPE ## _alloc_t *alloc)\
{\
  TYPE ## _alloc_delete(alloc);\
  TYPE ## _alloc_fill(alloc);\
}\
\
void TYPE ## _alloc_delete(struct TYPE ## _alloc_t *alloc)\
{\
  int i;\
  for (i = 0; i < alloc->alloc.i; i++)\
    TYPE ## _delete(&(alloc->data[i]));\
  if (alloc->alloc.previous)\
    {\
      TYPE ## _alloc_free((struct TYPE ## _alloc_t *)(alloc->alloc.previous));\
      alloc->alloc.previous = NULL;\
    }\
}\
\
void TYPE ## _alloc_free(struct TYPE ## _alloc_t *alloc)\
{\
  if (!alloc)\
    return;\
  TYPE ## _alloc_delete(alloc);\
  free(alloc);\
}

#endif
