#define ZASM_NAMESPACE_C 1

#include <zasm/namespace.h>
#include <zasm/memory.h>

struct namespace_t *namespace_new()
{
  struct namespace_t *val;

  ALLOCATE(struct namespace_t, val, 1);
  namespace_fill(val);
  return val;
}

void namespace_fill(struct namespace_t *ns)
{
  ns->local = NULL;
  ns->global = NULL;
}

void namespace_add_level(struct namespace_t *ns, struct ns_level_t *level)
{
  level->up = ns->local;
  ns->local = level;
  if (!(ns->global))
    ns->global = level;
}

struct ns_level_t *namespace_del_level(struct namespace_t *ns)
{
  struct ns_level_t *level;

  level = ns->local;
  if (!level)
    ERROR("unmatched #end_section or #pp_end_section\n");
  ns->local = level->up;
  if (!(ns->local))
    ns->global = NULL;
  return level;
}

struct hash_entry_t **namespace_lookup(struct namespace_t *ns, char *name, struct ns_level_t **lvl)
{
  struct ns_level_t *level;
  struct hash_entry_t **entry;
  
  for (level = ns->local; level; level = level->up)
    {
      entry = hash_table_lookup(&(level->hash_table), name);
      if (*entry)
	{
	  if (lvl)
	    *lvl = level;
	  return entry;
	}
    }
  if (lvl)
    *lvl = ns->local;
  return hash_table_lookup(&(ns->local->hash_table), name);
}

int namespace_is_empty(struct namespace_t *ns)
{
  if ((ns->global != NULL) || (ns->local != NULL))
    return 0;
  return 1;
}

void namespace_delete(struct namespace_t *ns)
{
  if ((ns->global != NULL) || (ns->local != NULL))
    {
      DEBUG("namespace is still in use\n");
      ERROR("internal error\n");
    }
}

void namespace_free(struct namespace_t *ns)
{
  namespace_delete(ns);
  free(ns);
}

struct ns_level_t *ns_level_new(struct namespace_t *parent)
{
  struct ns_level_t *val;
  
  ALLOCATE(struct ns_level_t, val, 1);
  ns_level_fill(val, parent);
  return val;
}

void ns_level_fill(struct ns_level_t *level, struct namespace_t *parent)
{
  hash_table_fill(&(level->hash_table));
  level->parent = parent;
  level->up = NULL;
}

void ns_level_delete(struct ns_level_t *level)
{
  hash_table_delete(&(level->hash_table));
}

void ns_level_free(struct ns_level_t *level)
{
  ns_level_delete(level);
  free(level);
}
