module Parser{-(processFile)-} where
    import Text.ParserCombinators.Parsec
    import Prototypes
    import Maybe

    type SParser a = CharParser Int a

    eSP :: SParser a -> String -> a
    eSP p s = case (runParser p 0 "" s) of
					Left err -> error ("parse error at "++show err)
					Right x -> x

-- normal functions
    processFile :: String -> String -> IO ()
    processFile src dest = do f <- readFile src
			      case (runParser file 0 src f) of 
					 Left err -> error ("parse error at " ++ show err)
					 Right x -> writeFile dest x

    concat2 :: [([a],[b])] -> ([a],[b])
    concat2 = c2.unzip
	where c2 :: ([[a]],[[b]]) -> ([a],[b])
	      c2 (l1, l2) = (concat l1, concat l2)

    makeStmts :: [(String,String)] -> String
    makeStmts = c.concat2
	where c :: ([a],[a]) -> [a]
	      c (x,y) = x++y

    makeInits :: [(String,String)] -> String
    makeInits = c.concat2
	where c :: ([a],[a]) -> [a]
	      c (x,y) = y++x

-- parsers
    nameString :: String -> SParser String
    nameString s = do {x <- string s; notFollowedBy (alphaNum<|>char '_'); return x}

    manyUntill :: SParser a -> SParser b -> SParser [a]
    manyUntill p end = many (do { notFollowedBy (do {try end; return ' '})
			       ; p
			       })

    manyUntill1 :: SParser a -> SParser b -> SParser [a]
    manyUntill1 p end = many1 (do { notFollowedBy (do {try end; return ' '})
				  ; p
				  })

    nsepBy :: Int -> SParser a -> SParser sep -> SParser [a]
    nsepBy 0 _ _ = do return []
    nsepBy 1 a _ = do x <- a
		      return [x]
    nsepBy n a s = do x <- a
		      s
		      y <- nsepBy (n-1) a s
		      return (x:y)

    pc_name :: SParser String
    pc_name = do n <- getState
		 setState (n+1)
		 return ("pc_parser_"++show n)

    pc_operator :: String -> Maybe Int -> Maybe Int -> Prototype -> SParser (String,PParser)
    pc_operator op nargs pargs proto = do try (nameString op)
					  as <- callArgs nargs pargs
					  name <- pc_name
					  return (makePrototype name proto as)

    pc_op_switch :: SParser (String,PParser)
    pc_op_switch = do try (nameString "pc_switch")
		      as <- callArgsMixed
		      name <- pc_name
		      return (makeSwitch name as)

    parser :: SParser (String,PParser)
    parser = between spaces spaces parserI

    parserI :: SParser (String,PParser)
    parserI = pc_operator "pc_token" (Just 1) (Just 0) pc_token
	     <|> pc_operator "pc_oneOf" Nothing (Just 0) pc_oneOf
	     <|> pc_operator "pc_noneOf" Nothing (Just 0) pc_noneOf
	     <|> pc_operator "pc_notFollowedByToken" (Just 1) (Just 0) pc_notFollowedByToken
	     <|> pc_operator "pc_tokenSequence" Nothing (Just 0) pc_tokenSequence
	     <|> pc_operator "pc_anyToken" (Just 0) (Just 0) pc_anyToken
	     <|> pc_operator "pc_empty" (Just 0) (Just 0) pc_empty
	     <|> pc_operator "pc_space" (Just 0) (Just 0) pc_space
	     <|> pc_operator "pc_spaces" (Just 0) (Just 0) pc_spaces
	     <|> pc_operator "pc_or" (Just 0) Nothing pc_or
	     <|> pc_op_switch
	     <|> pc_operator "pc_sequence" (Just 0) Nothing pc_sequence
	     <|> pc_operator "pc_try" (Just 0) (Just 1) pc_try
	     <|> pc_operator "pc_notFollowedBy" (Just 0) (Just 1) pc_notFollowedBy
	     <|> pc_operator "pc_maybe" (Just 0) (Just 1) pc_maybe
	     <|> pc_operator "pc_many" (Just 0) (Just 1) pc_many
	     <|> pc_operator "pc_skipMany" (Just 0) (Just 1) pc_skipMany
	     <|> pc_operator "pc_many1" (Just 0) (Just 1) pc_many1
	     <|> pc_operator "pc_count" (Just 1) (Just 1) pc_count
	     <|> pc_operator "pc_manyTill" (Just 0) (Just 2) pc_manyTill
	     <|> pc_operator "pc_manyTill1" (Just 0) (Just 2) pc_manyTill1
	     <|> pc_operator "pc_sepBy" (Just 0) (Just 2) pc_sepBy
	     <|> pc_operator "pc_sepBy1" (Just 0) (Just 2) pc_sepBy1
	     <|> pc_operator "pc_sepByN" (Just 0) (Just 2) pc_sepByN
	     <|> pc_operator "pc_endBy" (Just 0) (Just 2) pc_endBy
	     <|> pc_operator "pc_endBy1" (Just 0) (Just 2) pc_endBy1
	     <|> pc_operator "pc_endByN" (Just 0) (Just 2) pc_endByN
	     <|> pc_operator "pc_begBy" (Just 0) (Just 2) pc_begBy
	     <|> pc_operator "pc_begBy1" (Just 0) (Just 2) pc_begBy1
	     <|> pc_operator "pc_begByN" (Just 0) (Just 2) pc_begByN
	     <|> pc_operator "pc_sepEndBy" (Just 0) (Just 2) pc_sepEndBy
	     <|> pc_operator "pc_sepEndBy1" (Just 0) (Just 2) pc_sepEndBy1
	     <|> pc_operator "pc_sepEndByN" (Just 0) (Just 2) pc_sepEndByN
	     <|> pc_operator "pc_between" (Just 0) (Just 3) pc_between
	     <|> pc_operator "pc_betweenTokens" (Just 2) (Just 1) pc_betweenTokens
	     <|> pc_operator "pc_parens" (Just 0) (Just 1) pc_parens
	     <|> pc_operator "pc_braces" (Just 0) (Just 1) pc_braces
	     <|> pc_operator "pc_brackets" (Just 0) (Just 1) pc_brackets
	     <|> pc_operator "pc_squares" (Just 0) (Just 1) pc_squares
	     <|> do {x <- getName; as <- callArgs Nothing (Just 0); return (makeName x as);}
	     <|> between (char '{') (char '}') parser_block

    callArgsMixed :: SParser [(String,(String,PParser))]
    callArgsMixed = do spaces
		       as <- between (char '(') (char ')') (sepBy (do {t <- expr
								      ; char ','
								      ; p<-parser
								      ; return (t,p)}) (char ',')) <|> noArg
		       return as

    callArgs :: Maybe Int -> Maybe Int -> SParser [Either String (String,PParser)]
    callArgs (Nothing) (Just 0) = do spaces
				     as <- (between (char '(') (char ')') (sepBy expr (char ',')) <|> noArg)
				     return (map Left as)
    callArgs (Just 0) (Nothing) = do spaces
				     as <- (between (char '(') (char ')') (sepBy parser (char ',')) <|> noArg)
				     return (map Right as)
    callArgs (Just 0) (Just 0) = do spaces
				    between (char '(') (char ')') (many space) <|> noArg
				    return []
    callArgs (Just n) (Just 0) = do spaces
				    as <- between (char '(') (char ')') (nsepBy n expr (char ','))
				    return (map Left as)
    callArgs (Just 0) (Just m) = do spaces
				    as <- between (char '(') (char ')') (nsepBy m parser (char ','))
				    return (map Right as)
    callArgs (Just n) (Just m) = do spaces
				    between (char '(') (char ')') (do{as <- nsepBy n expr (char ',');
								      char ','; 
								      bs <- nsepBy m parser (char ',');
								      return ((map Left as) ++ (map Right bs));
								     }
								  )
    callArgs (Just n) (Nothing) = do spaces
				     between (char '(') (char ')') (do{as <- nsepBy n expr (char ',');
								       char ','; 
								       bs <- sepBy parser (char ',');
								       return ((map Left as) ++ (map Right bs));
								      }
								   )

    -- for now, no parsers are allowed in expressions
    expr :: SParser String
    expr = do x <- many (many1 (noneOf "(,)")<|>allBetween (string "(") (string ")"))
	      return (concat x)

    noArg :: SParser [a]
    noArg = do return []



    allBetween :: SParser String -> SParser String -> SParser String
    allBetween b e = do x <- try b
			y <- many ((allBetween b e) <|> (manyUntill1 anyToken (try b<|>try e)))
			z <- e
			return (x++concat y++z)

    cComments :: SParser String -> SParser String -> SParser String
    cComments b e = do x <- try b
		       y <- manyUntill anyChar (try e)
		       z <- e
		       return (x++y++z)

    getName :: SParser String
    getName = many1 (alphaNum<|>char '_')

    -- call a parser on one of the following ways:
    --   var = pc(parser);
    --   var <- parser;
    --   pc(parser);
    parser_block :: SParser (String, PParser)
    parser_block = do cs <- many (parser_call<|>parser_stmt)
		      name <- pc_name
		      return (makeBlock name cs)

    parser_stmt :: SParser (String, String)
    parser_stmt = do { x <- many (noneOf "{};")
		     ; (def,st) <- do { char ';'
				      ; y <- many space
				      ; return ("",';':y)
				      }
		       <|> do { l <- between (char '{') (char '}') (many (parser_call<|>parser_stmt))
			      ; y <- many space
			      ; return (concat2 ((("","{"):l)++[("",'}':y)]))
			      }
		     ; return (def,x++st)
		     }

    parser_call :: SParser (String, String)
    parser_call = try (do {spaces
			  ; name <- getName
			  ; spaces
			  ; char '='
			  ; (pinit,pp) <- parser_pc
			  ; spaces
			  ; char ';'
			  ; spaces
			  ; return (pinit,pcalln name pp)
			  }
		      )
		  <|> try (do {spaces
			      ; nameString "return"
			      ; (pinit,pp) <- parser_pc
			      ; spaces
			      ; char ';'
			      ; spaces
			      ; return (pinit,pcall pp++"return pc_retval;\n")
			      }
			  )
		  <|> try (do {spaces
			      ; name <- getName
			      ; spaces
			      ; string "<-"
			      ; spaces
			      ; (pinit,pp) <- parser
			      ; spaces
			      ; char ';'
			      ; spaces
			      ; return (pinit,pcalln name pp)
			      }
			  )
		  <|> try (do {(pinit,pp) <- parser_pc
			      ; spaces
			      ; char ';'
			      ; spaces
			      ; return (pinit,pcall pp)
			      }
			  )
	where pcalln :: String -> PParser -> String
	      pcalln name pp = 
		  ("  "++name++"="++callPParser pp++";"<>
		   "  if (("++name++".type == TOK_PC_FAIL)||("++name++".type == TOK_PC_FAIL1))"<>
		   "  {"<>
		   "    if (pc_called)"<>
		   "      "++name++".type = TOK_PC_FAIL;"<>
		   "    return "++name++";"<>
		   "  }"<>
		   "  pc_called = 1;\n")
	      pcall :: PParser -> String
	      pcall = pcalln "pc_retval"

    parser_pc :: SParser (String, PParser)
    parser_pc = do {spaces; nameString "pc"; spaces; between (char '(') (char ')') parser;}

    file :: SParser String
    file = do inits <- manyUntill stmt (string "__PARSER_DEFS")
	      string "__PARSER_DEFS"
	      stmts <- many stmt
	      return (makeInits inits ++ makeStmts stmts)

    stmt :: SParser (String,String)
    stmt = do {x <- cComments (string "/*") (string "*/"); return ("",x);}
	   <|> do {x <- try (string "//"); y <- many (noneOf "\n"); char '\n'; return ("",x++y++"\n")}
--	   <|> do {char '#'; y <- many (noneOf "\n"); char '\n'; return ("",('#':y)++"\n")}
--	   <|> do {x <- many1 space; return ("",x)}
	   <|> do {try (nameString "__PARSER")
		  ;(name,args,defs,pp) <- parser_def
		  ;return (fromMaybe "" defs++makeDefinition name args pp,"")
		  }
	   <|> do {try (nameString "__INLINE_PARSER")
		  ;(name,args,defs,pp) <- parser_def
		  ;return (fromMaybe "" defs++"static inline "++makeDefinition name args pp,"")
		  }
	   <|> do {try (nameString "pc_execute")
		  ;spaces
		  ;(defs,pp,feeder,eof) <- between (char '(') (char ')') 
		   (do {(d1,p1)<-parser;
			char ',';
			f1<-expr;
			char ',';
			e1<-expr;
			return (d1,p1,f1,e1)})
		  ;name <- pc_name
		  ;return (defs++pc_execute name pp,name++"("++feeder++","++eof++")")
		  }
	   <|> do { x <- manyUntill1 anyToken (try (string "/*")<|>
					       try (string "//")<|>
					       try (nameString "__PARSER")<|>
					       try (nameString "__INLINE_PARSER")<|>
					       try (nameString "__PARSER_DEFS")<|>
					       try (nameString "pc_execute"))
		  ; return ("",x)
		  }

    parser_def :: SParser (String,[String],Maybe String,Maybe PParser)
    parser_def = do spaces
		    name <- getName
		    spaces
		    args <- parser_args
		    spaces
		    (def,pp) <- do { char '='
				   ; (def1,pp1) <- parser
				   ; char ';'
				   ; return ((Just def1),(Just pp1))
				   }
				 <|> do { char ';'
					; return (Nothing,Nothing)
					}
		    return (name,args,def,pp)

    parser_args :: SParser [String]
    parser_args = between (char '(') (char ')') (sepBy expr (char ',')) <|> noArg
