module Prototypes where
    import List

    data PParser = PParser String [String]
    type Prototype = String -> [PParser] -> Int -> String

    (<>) :: String -> String -> String
    a <> b = a ++ ('\n':b)

    pc_execute :: String -> PParser -> String
    pc_execute name (PParser pname args) =
	(""<>
	 "static inline struct token_t "++name++"(struct feeder_t *input, int *endOfFile)"<>
         "{"<>
         "  struct token_t token, retval;"<>
	 "  token = feeder_get(input);"<>
	 "  retval = "++pname++"(input,&token"++concatMap (',':) args++");"<>
	 "  if ((retval.type == TOK_PC_FAIL1) || (retval.type == TOK_PC_FAIL))"<>
	 "    T_ERROR(token, \"syntax error\\n\");"<>
	 "  *endOfFile = 0;"<>
	 "  if (token.type != TOK_END_OF_FILE)"<>
	 "    feeder_unget(input, token);"<>
	 "  else"<>
	 "    *endOfFile = 1;"<>
	 "  return retval;"<>
	 "}\n")

    makeDefinition :: String -> [String] -> Maybe PParser -> String
    makeDefinition name args (Just pp) = 
	(""<>
	 "struct token_t "++name++"(struct feeder_t *input, struct token_t *token"++concatMap (',':) args++")"<>
	 "{"<>
	 "  return "++callPParser pp++";"<>
	 "}\n")
    makeDefinition name args Nothing   = 
	("struct token_t "++name++"(struct feeder_t *input, struct token_t *token"++concatMap (',':) args++");\n")

    makeName :: String -> [Either String (String, PParser)] -> (String, PParser)
    makeName name args = ("", PParser name (map l args))
	where l :: Either a b -> a
	      l (Left x) = x

    makeBlock :: String -> [(String,String)] -> (String,PParser)
    makeBlock name all = 
	(concat pdefs++definePParser name []++
	 "{"<>
	 "  struct token_t pc_retval;"<>
	 "  int pc_called = 0;"<>
	 concat stmts<>
	 "}\n"
	, PParser name [])
	where (pdefs,stmts) = unzip all

    makeSwitch :: String -> [(String ,(String, PParser))] -> (String, PParser)
    makeSwitch name args = (concat defs++pc_switch name cargs, PParser name [])
	where f (lbl,(def,pp)) = (def,(lbl,pp))
	      (defs,cargs) = unzip (map f args)

    makePrototype :: String -> Prototype -> [Either String (String, PParser)] -> (String, PParser)
    makePrototype name proto args = (concat pdefs++proto name pargs (length nargs), PParser name nargs)
	where (nargs,ppargs) = partEither args
	      (pdefs,pargs) = unzip ppargs
	      partEither :: [Either a b] -> ([a],[b])
	      partEither [] = ([],[])
	      partEither (Left x:es) = (x:xs,ys)
		  where (xs,ys) = partEither es
	      partEither (Right y:es) = (xs,y:ys)
		  where (xs,ys) = partEither es

    callPParser :: PParser -> String
    callPParser (PParser name args) = name++"(input,token"++concatMap (',':) args++")"

    definePParser :: String -> [String] -> String
    definePParser name args = 
	(""<>
	 "static inline struct token_t "++name++"(struct feeder_t *input, struct token_t *token"++concatMap (',':) args++")"
	)

-- prototypes -----------------------------------------------------------------------------------------------------------------
  -- primitive parsers
    pc_token :: Prototype
    pc_token name []  _ =
	(definePParser name ["int type"]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result.position.file = NULL;"<>
	 "  result.position.original = NULL;"<>
	 "  if (token->type == type)"<>
	 "  {"<>
	 "    result = *token;"<>
	 "    *token = feeder_get(input);"<>
	 "  }"<>
	 "  else result.type = TOK_PC_FAIL1;"<>
	 "  return result;"<>
	 "}\n")
    
    pc_oneOf :: Prototype
    pc_oneOf name [] 0 = error "oneOf expects at least 1 argument\n"
    pc_oneOf name [] n =
	(definePParser name ["int type"++show i| i <- [1..n]]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result.position.file = NULL;"<>
	 "  result.position.original = NULL;"<>
	 "  if ((token->type == type1)"++concat ["||(token->type == type"++show i++")"| i <- [2..n]]++")"<>
	 "  {"<>
	 "    result = *token;"<>
	 "    *token = feeder_get(input);"<>
	 "  }"<>
	 "  else"<>
	 "    result.type = TOK_PC_FAIL1;"<>
	 "  return result;"<>
	 "}\n")

    pc_noneOf :: Prototype
    pc_noneOf name [] n =
	(definePParser name ["int type"++show i| i <- [1..n]]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result.position.file = NULL;"<>
	 "  result.position.original = NULL;"<>
	 "  if ((token->type == type1)"++concat ["||(token->type == type"++show i++")"| i <- [2..n]]++")"<>
	 "    result.type = TOK_PC_FAIL1;"<>
	 "  else"<>
	 "  {"<>
	 "    result = *token;"<>
	 "    *token = feeder_get(input);"<>
	 "  }"<>
	 "  return result;"<>
	 "}\n")

    pc_notFollowedByToken :: Prototype
    pc_notFollowedByToken name [] _ =
	(definePParser name ["int type"]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result.position.file = NULL;"<>
	 "  result.position.original = NULL;"<>
	 "  if (token->type == type)"<>
	 "    result.type = TOK_PC_FAIL1;"<>
	 "  else"<>
	 "    result.type = TOK_PC_SUCCESS;"<>
	 "  return result;"<>
	 "}\n")

    pc_tokenSequence :: Prototype
    pc_tokenSequence name [] 0 = pc_empty name [] 0
    pc_tokenSequence name [] n = 
	(definePParser name ["int type"++show i| i <- [1..n]]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  list = list_new();"<>
	 "  result.position.file = NULL;"<>
	 "  result.position.original = NULL;"<>
	 "  if (token->type != type1)"<>
	 "  {"<>
	 "    result.type = TOK_PC_FAIL;"<>
	 "    list_free(list);"<>
	 "    return result;"<>
	 "  }"<>
	 "  list_add_end(list, *token);"<>
	 "  *token = feeder_get(input);"<>
	 concat ["  if (token->type != type"++show i++")"<>
		 "  {"<>
		 "    result.type = TOK_PC_FAIL;"<>
		 "    list_free(list);"<>
		 "    return result;"<>
		 "  }"<>
		 "  list_add_end(list, *token);"<>
		 "  *token = feeder_get(input);\n" | i <- [2..n]]++
	 "  result.type = TOK_LIST;"<>
	 "  result.value.t_struct_list_t_p = list;"<>
	 "  return result;"<>
	 "}\n")

    pc_anyToken :: Prototype
    pc_anyToken name [] _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result = *token;"<>
	 "  *token = feeder_get(input);"<>
	 "  return result;"<>
	 "}\n")

    pc_empty :: Prototype
    pc_empty name [] _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result.type = TOK_PC_SUCCESS;"<>
	 "  return result;"<>
	 "}\n")

    pc_space :: Prototype
    pc_space name [] _ = 
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  if (token->type == TOK_SPACE)"<>
	 "  {"<>
	 "    result = *token;"<>
	 "    *token = feeder_get(input);"<>
	 "  }"<>
	 "  else result.type = TOK_PC_FAIL1;"<>
	 "  return result;"<>
	 "}\n")

    pc_spaces :: Prototype
    pc_spaces name [] _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  while (token->type == TOK_SPACE)"<>
	 "    *token = feeder_get(input);"<>
	 "  result.type = TOK_PC_SUCCESS;"<>
	 "  return result;"<>
	 "}\n")

  -- alternatives
    pc_or :: Prototype
    pc_or name ps _ = 
	(definePParser name [] <>
	 "{"<>
	 "  struct token_t result;"<>
	 concatMap (\p->
		    ("  result="++callPParser p++ ";"<>
		     "  if (result.type != TOK_PC_FAIL1)"<>
		     "    return result;")) ps <>
	 "  return result;"<>
	 "}\n")

    pc_switch :: String -> [(String,PParser)] -> String
    pc_switch name ps =
	(definePParser name [] <>
	 "{"<>
	 "  struct token_t result;"<>
	 "  switch (token->type)"<>
	 "  {"<>
	 concat ["  case "++i++":"<>
		 "    *token = feeder_get(input);"<>
		 "    result = "++callPParser p++";"<>
		 "    if (result.type == TOK_PC_FAIL1)"<>
		 "      result.type = TOK_PC_FAIL;"<>
		 "    return result;\n" | (i,p) <- ps]++
	 "  default:"<>
	 "    result.type = TOK_PC_FAIL1;"<>
	 "    return result;"<>
	 "  }"<>
	 "}\n")

  -- misc
    pc_sequence :: Prototype
    pc_sequence _ [] _ = error "a pc_sequence should have at least 1 argument\n"
    pc_sequence name (p:ps) _ =
	(definePParser name [] <>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  result = "++callPParser p++";"<>
	 "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    return result;"<>
	 "  list = list_new();"<>
	 "  list_add_end(list, result);"<>
	 concat ["  result = "++callPParser pp++";"<>
		 "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
		 "  {"<>
		 "    result.type = TOK_PC_FAIL;"<>
		 "    list_free(list);"<>
		 "    return result;"<>
		 "  }"<>
		 "  list_add_end(list, result);\n"| pp <- ps]++
	 "  result.type = TOK_LIST;"<>
	 "  result.value.t_struct_list_t_p = list;"<>
	 "  return result;"<>
	 "}\n")

    pc_try :: Prototype
    pc_try name (p:[])  _ =
	(definePParser name [] <>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct token_t backup;"<>
	 "  backup = *token;"<>
	 "  feeder_backup(input);"<>
	 "  result = "++callPParser p++";"<>
	 "  if (result.type == TOK_PC_FAIL)"<>
	 "  {"<>
	 "     feeder_restore(input);"<>
	 "     *token = backup;"<>
	 "     result.type = TOK_PC_FAIL1;"<>
	 "  }"<>
	 "  else"<>
	 "    feeder_stop_backup(input);"<>
	 "  return result;"<>
	 "}\n")

    pc_notFollowedBy :: Prototype
    pc_notFollowedBy name (p:[]) _ =
	(definePParser name [] <>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct token_t backup;"<>
	 "  backup = *token;"<>
	 "  feeder_backup(input);"<>
	 "  result = "++callPParser p++";"<>
	 "  feeder_restore(input);"<>
	 "  *token = backup;"<>
	 "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_FAIL1))"<>
	 "    result.type = TOK_PC_SUCCESS;"<>
	 "  else"<>
	 "    result.type = TOK_PC_FAIL1;"<>
	 "  return result;"<>
	 "}\n")

  -- repetitors
    pc_maybe :: Prototype
    pc_maybe name (p:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  result = "++callPParser p++";"<>
	 "  if (result.type == TOK_PC_FAIL1)"<>
	 "    result.type = TOK_PC_SUCCESS;"<>
	 "  return result;"<>
	 "}\n")

    pc_many :: Prototype
    pc_many name (p:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  list = list_new();"<>
	 "  while (1)"<>
	 "  {"<>
	 "    result = "++callPParser p++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list, result);"<>
	 "  }"<>
	 "}\n")

    pc_skipMany :: Prototype
    pc_skipMany name (p:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  while (1)"<>
	 "  {"<>
	 "    result = "++callPParser p++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      return result;"<>
	 "    }"<>
	 "    if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      result.type = TOK_PC_SUCCESS;"<>
	 "      return result;"<>
	 "    }"<>
	 "  }"<>
	 "}\n")

    pc_many1 :: Prototype
    pc_many1 name (p:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  result = "++callPParser p++";"<>
	 "  if ((result.type == TOK_PC_FAIL) || (result.type == TOK_PC_FAIL1))"<>
	 "    return result;"<>
	 "  list = list_new();"<>
	 "  list_add_end(list, result);"<>
	 "  while (1)"<>
	 "  {"<>
	 "    result = "++callPParser p++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list, result);"<>
	 "  }"<>
	 "}\n")

    pc_count :: Prototype
    pc_count name (p:[]) _ =
	(definePParser name ["int l"]<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  int i;"<>
	 "  struct list_t *list;"<>
	 "  list = list_new();"<>
	 "  for (i = 0; i < l; i++)"<>
	 "  {"<>
	 "    result = "++callPParser p++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    else if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      if (i)"<>
	 "        result.type = TOK_PC_FAIL;"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list, result);"<>
	 "  }"<>
	 "  result.type = TOK_LIST;"<>
	 "  result.value.t_struct_list_t_p = list;"<>
	 "  return result;"<>
	 "}\n")

    pc_manyTill :: Prototype
    pc_manyTill name (p:end:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct token_t backup;"<>
	 "  struct list_t *list;"<>
	 "  int i;"<>
	 "  list = list_new();"<>
	 "  for (i = 0; 1; i=1)"<>
	 "  {"<>
	 "    backup = *token;"<>
	 "    feeder_backup(input);"<>
	 "    result = "++callPParser end++";"<>
	 "    feeder_restore(input);"<>
	 "    *token = backup;"<>
	 "    if ((result.type != TOK_PC_FAIL)&&(result.type != TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    result = "++callPParser p++";"<>
	 "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      if (i)"<>
	 "        result.type = TOK_PC_FAIL;"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list, result);"<>
	 "  }"<>
	 "}\n")

    pc_manyTill1 :: Prototype
    pc_manyTill1 name (p:end:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct token_t backup;"<>
	 "  struct list_t *list;"<>
	 "  int i;"<>
	 "  list = list_new();"<>
	 "  for (i = 0; 1; i=1)"<>
	 "  {"<>
	 "    backup = *token;"<>
	 "    feeder_backup(input);"<>
	 "    result = "++callPParser end++";"<>
	 "    feeder_restore(input);"<>
	 "    *token = backup;"<>
	 "    if ((result.type != TOK_PC_FAIL)&&(result.type != TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      if (!i)"<>
	 "      {"<>
	 "        result.type = TOK_PC_FAIL1;"<>
	 "        return result;"<>
	 "      }"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    result = "++callPParser p++";"<>
	 "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      if (i)"<>
	 "        result.type = TOK_PC_FAIL;"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list, result);"<>
	 "  }"<>
	 "}\n")

  -- common syntactical structures
  --  lists
  --    separated
    pc_sepBy :: Prototype
    pc_sepBy name (p:sep:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  list = list_new();"<>
	 "  result = "++callPParser p++";"<>
	 "  if (result.type == TOK_PC_FAIL)"<>
	 "  {"<>
	 "    list_free(list);"<>
	 "    return result;"<>
	 "  }"<>
	 "  if (result.type == TOK_PC_FAIL1)"<>
	 "  {"<>
	 "    result.type = TOK_LIST;"<>
	 "    result.value.t_struct_list_t_p = list;"<>
	 "    return result;"<>
	 "  }"<>
	 "  list_add_end(list, result);"<>
	 "  while (1)"<>
	 "  {"<>
	 "    result = "++callPParser sep++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    result = "++callPParser p++";"<>
	 "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      result.type = TOK_PC_FAIL;"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list,result);"<>
	 "  }"<>
	 "}\n")
	
    pc_sepBy1 :: Prototype
    pc_sepBy1 name (p:sep:[]) _ =
	(definePParser name []<>
	 "{"<>
	 "  struct token_t result;"<>
	 "  struct list_t *list;"<>
	 "  result = "++callPParser p++";"<>
	 "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    return result;"<>
	 "  list = list_new();"<>
	 "  list_add_end(list, result);"<>
	 "  while (1)"<>
	 "  {"<>
	 "    result = "++callPParser sep++";"<>
	 "    if (result.type == TOK_PC_FAIL)"<>
	 "    {"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    if (result.type == TOK_PC_FAIL1)"<>
	 "    {"<>
	 "      result.type = TOK_LIST;"<>
	 "      result.value.t_struct_list_t_p = list;"<>
	 "      return result;"<>
	 "    }"<>
	 "    result = "++callPParser p++";"<>
	 "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
	 "    {"<>
	 "      result.type = TOK_PC_FAIL;"<>
	 "      list_free(list);"<>
	 "      return result;"<>
	 "    }"<>
	 "    list_add_end(list,result);"<>
	 "  }"<>
	 "}\n")

    pc_sepByN :: Prototype
    pc_sepByN name (p:sep:[]) _ =
      (definePParser name ["int n"]<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  int i;"<>
       "  list = list_new();"<>
       "  if (n == 0)"<>
       "  {"<>
       "    result.type = TOK_LIST;"<>
       "    result.value.t_struct_list_t_p = list;"<>
       "    return result;"<>
       "  }"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    list_free(list);"<>
       "    return result;"<>
       "  }"<>
       "  list = list_new();"<>
       "  list_add_end(list, result);"<>
       "  for (i = 1; i < n; i++)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "  result.type = TOK_LIST;"<>
       "  result.value.t_struct_list_t_p = list;"<>
       "  return result;"<>
       "}\n")

  --    ended
    pc_endBy :: Prototype
    pc_endBy name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  list = list_new();"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser p++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "  }"<>
       "}\n")

    pc_endBy1 :: Prototype
    pc_endBy1 name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    return result;"<>
       "  list = list_new();"<>
       "  list_add_end(list,result);"<>
       "  result = "++callPParser sep++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL;"<>
       "    list_free(list);"<>
       "    return result;"<>
       "  }"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser p++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "  }"<>
       "}\n")

    pc_endByN :: Prototype
    pc_endByN name (p:sep:[]) _ =
      (definePParser name ["int n"]<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  int i;"<>
       "  list = list_new();"<>
       "  for (i = 0; i < n; i++)"<>
       "  {"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      if (i)"<>
       "        result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "  }"<>
       "  result.type = TOK_LIST;"<>
       "  result.value.t_struct_list_t_p = list;"<>
       "  return result;"<>
       "}\n")

  --    begun
    pc_begBy :: Prototype
    pc_begBy name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  list = list_new();"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "}\n")

    pc_begBy1 :: Prototype
    pc_begBy1 name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  result = "++callPParser sep++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    return result;"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL;"<>
       "    return result;"<>
       "  }"<>
       "  list = list_new();"<>
       "  list_add_end(list,result);"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "}\n")

    pc_begByN :: Prototype
    pc_begByN name (p:sep:[]) _ =
      (definePParser name ["int n"]<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  int i;"<>
       "  list = list_new();"<>
       "  for (i = 0; i < n; i++)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      if (i)"<>
       "        result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "  result.type = TOK_LIST;"<>
       "  result.value.t_struct_list_t_p = list;"<>
       "  return result;"<>
       "}\n")

  --    separated or ended
    pc_sepEndBy :: Prototype
    pc_sepEndBy name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  list = list_new();"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser p++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "    result = "++callPParser sep++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "  }"<>
       "}\n")

    pc_sepEndBy1 :: Prototype
    pc_sepEndBy1 name (p:sep:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    return result;"<>
       "  list = list_new();"<>
       "  list_add_end(list, result);"<>
       "  while (1)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if (result.type == TOK_PC_FAIL)"<>
       "    {"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    if (result.type == TOK_PC_FAIL1)"<>
       "    {"<>
       "      result.type = TOK_LIST;"<>
       "      result.value.t_struct_list_t_p = list;"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "}\n")

    pc_sepEndByN :: Prototype
    pc_sepEndByN name (p:sep:[]) _ =
      (definePParser name ["int n"]<>
       "{"<>
       "  struct token_t result;"<>
       "  struct list_t *list;"<>
       "  int i;"<>
       "  list = list_new();"<>
       "  if (n == 0)"<>
       "  {"<>
       "    result.type = TOK_LIST;"<>
       "    result.value.t_struct_list_t_p = list;"<>
       "    return result;"<>
       "  }"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    list_free(list);"<>
       "    return result;"<>
       "  }"<>
       "  list = list_new();"<>
       "  list_add_end(list, result);"<>
       "  for (i = 1; i < n; i++)"<>
       "  {"<>
       "    result = "++callPParser sep++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    result = "++callPParser p++";"<>
       "    if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    {"<>
       "      result.type = TOK_PC_FAIL;"<>
       "      list_free(list);"<>
       "      return result;"<>
       "    }"<>
       "    list_add_end(list,result);"<>
       "  }"<>
       "  result = "++callPParser sep++";"<>
       "  if (result.type == TOK_PC_FAIL)"<>
       "  {"<>
       "    list_free(list);"<>
       "    return result;"<>
       "  }"<>
       "  result.type = TOK_LIST;"<>
       "  result.value.t_struct_list_t_p = list;"<>
       "  return result;"<>
       "}\n")

  --  between and friends
    pc_between :: Prototype
    pc_between name (b:e:p:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result, retval;"<>
       "  result = "++callPParser b++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "    return result;"<>
       "  retval = "++callPParser p++";"<>
       "  if ((retval.type == TOK_PC_FAIL)||(retval.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    retval.type = TOK_PC_FAIL;"<>
       "    return retval;"<>
       "  }"<>
       "  result = "++callPParser e++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1))"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL;"<>
       "    return result;"<>
       "  }"<>
       "  return retval;"<>
       "}\n")

    pc_betweenTokens :: Prototype
    pc_betweenTokens name (p:[]) _ =
      (definePParser name ["int begin","int end"]<>
       "{"<>
       "  struct token_t result;"<>
       "  if (token->type != begin)"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL1;"<>
       "    return result;"<>
       "  }"<>
       "  *token = feeder_get(input);"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1)||(token->type != end))"<>
       "    result.type = TOK_PC_FAIL;"<>
       "  *token = feeder_get(input);"<>
       "  return result;"<>
       "}\n")
  -- ()
    pc_parens :: Prototype
    pc_parens name (p:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  if (token->type != TOK_PAREN_OPEN)"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL1;"<>
       "    return result;"<>
       "  }"<>
       "  *token = feeder_get(input);"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1)||(token->type != TOK_PAREN_CLOSE))"<>
       "    result.type = TOK_PC_FAIL;"<>
       "  *token = feeder_get(input);"<>
       "  return result;"<>
       "}\n")
  -- {}
    pc_braces :: Prototype
    pc_braces name (p:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  if (token->type != TOK_BRACE_OPEN)"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL1;"<>
       "    return result;"<>
       "  }"<>
       "  *token = feeder_get(input);"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1)||(token->type != TOK_BRACE_CLOSE))"<>
       "    result.type = TOK_PC_FAIL;"<>
       "  *token = feeder_get(input);"<>
       "  return result;"<>
       "}\n")
  -- <>
    pc_brackets :: Prototype
    pc_brackets name (p:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  if (token->type != TOK_BRACKET_OPEN)"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL1;"<>
       "    return result;"<>
       "  }"<>
       "  *token = feeder_get(input);"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1)||(token->type != TOK_BRACKET_CLOSE))"<>
       "    result.type = TOK_PC_FAIL;"<>
       "  *token = feeder_get(input);"<>
       "  return result;"<>
       "}\n")
  -- []
    pc_squares :: Prototype
    pc_squares name (p:[]) _ =
      (definePParser name []<>
       "{"<>
       "  struct token_t result;"<>
       "  if (token->type != TOK_SQUARE_OPEN)"<>
       "  {"<>
       "    result.type = TOK_PC_FAIL1;"<>
       "    return result;"<>
       "  }"<>
       "  *token = feeder_get(input);"<>
       "  result = "++callPParser p++";"<>
       "  if ((result.type == TOK_PC_FAIL)||(result.type == TOK_PC_FAIL1)||(token->type != TOK_SQUARE_CLOSE))"<>
       "    result.type = TOK_PC_FAIL;"<>
       "  *token = feeder_get(input);"<>
       "  return result;"<>
       "}\n")
