#define ZASM_ERRORS_C 1

#include <zasm/errors.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static FILE *error_file = NULL;
static FILE *warning_file = NULL;
static FILE *debug_file = NULL;

void error_init(int *argc_p, char ***argv_p)
{
  int argc = *argc_p;
  char **argv = *argv_p;
  char **ret_argv;
  int ret_argc = argc;
  int i;

  // set sensible defaults
  error_file = stderr;
  warning_file = stderr;
  debug_file = stderr;

  for (i = 1; i < argc; i++)
    {
      if ((0==strcasecmp("--debug", argv[i])) ||
          (0==strcasecmp("--warning", argv[i])) ||
          (0==strcasecmp("--error", argv[i])) ||
          (0==strcasecmp("--nodebug", argv[i])) ||
          (0==strcasecmp("--nowarning", argv[i])) ||
          (0==strcasecmp("--noerror", argv[i])))
        ret_argc--;
      else if ((0==strcasecmp("--debug-file", argv[i])) ||
	       (0==strcasecmp("--warning-file", argv[i])) ||
	       (0==strcasecmp("--error-file", argv[i])))
        {
          ret_argc-=2;
          i++;
        }
    }

  if (ret_argc == argc)
    return;
  
  ret_argv = (char **)malloc(sizeof(char *)*ret_argc);
  if (!ret_argv)
    {
      fprintf(stderr, "allocation error\n");
      exit(1);
    }

  ret_argv[0] = argv[0];

  for (i = 1,ret_argc = 1; i < argc; i++)
    {
      if (0==strcasecmp("--debug", argv[i]))
        debug_file = stderr;
      else if (0==strcasecmp("--warning", argv[i]))
        warning_file = stderr;
      else if (0==strcasecmp("--error", argv[i]))
        error_file = stderr;
      else if (0==strcasecmp("--nodebug", argv[i]))
        debug_file = NULL;
      else if (0==strcasecmp("--nowarning", argv[i]))
        warning_file = NULL;
      else if (0==strcasecmp("--noerror", argv[i]))
        error_file = NULL;
      else if (0==strcasecmp("--debug-file", argv[i]))
        {
          i++;
          if (i >= argc)
            {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  if (0==strcmp("-", argv[i]))
	    debug_file = stdout;
	  else
	    debug_file = fopen(argv[i], "w");
        }
      else if (0==strcasecmp("--warning-file", argv[i]))
        {
          i++;
          if (i >= argc)
            {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  if (0==strcmp("-", argv[i]))
	    warning_file = stdout;
	  else
	    warning_file = fopen(argv[i], "w");
        }
      else if (0==strcasecmp("--error-file", argv[i]))
        {
          i++;
          if (i >= argc)
            {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  if (0==strcmp("-", argv[i]))
	    error_file = stdout;
	  else
	    error_file = fopen(argv[i], "w");
        }
      else
        {
          ret_argv[ret_argc] = argv[i];
          ret_argc++;
        }
    }
  *argc_p = ret_argc;
  *argv_p = ret_argv;
}

static void error_print_location(FILE *file, char *type, struct token_t token)
{
  fprintf(file, "%s:\n", type);
  fprintf(file, "  at token \"");
  token_print_nowarning(token, file);
  fprintf(file, "\" in \"%s\" at lines %i-%i\n",
	  token.position.file, token.position.line_begin, token.position.line_end);
  while (token.position.original)
    {
      token = *(token.position.original);
      fprintf(file, "  in substitution of token \"");
      token_print_nowarning(token, file);
      fprintf(file, "\" in \"%s\" at lines %i-%i\n",
	      token.position.file, token.position.line_begin, token.position.line_end);
    }
}

void error_print_error_token(struct token_t token, char *format, ...)
{
  va_list ap;
  
  if (error_file)
    {
      error_print_location(error_file, "error", token);
      fprintf(error_file, "    ");
      va_start(ap, format);
      vfprintf(error_file, format, ap);
      va_end(ap);
    }
  exit(1);
}

void error_print_warning_token(struct token_t token, char *format, ...)
{
  va_list ap;
  
  if (warning_file)
    {
      error_print_location(warning_file, "warning", token);
      fprintf(error_file, "    ");
      va_start(ap, format);
      vfprintf(warning_file, format, ap);
      va_end(ap);
    }
}

void error_print_debug_token(struct token_t token, char *format, ...)
{
  va_list ap;
  
  if (debug_file)
    {
      error_print_location(warning_file, "debug", token);
      fprintf(error_file, "    ");
      va_start(ap, format);
      vfprintf(debug_file, format, ap);
      va_end(ap);
    }
}

void error_print_error(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  
  if (error_file)
    {
      fprintf(error_file, "error: ");
      vfprintf(error_file, format, ap);
    }

  va_end(ap);

  exit(1);
}

void error_print_warning(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  
  if (warning_file)
    {
      fprintf(warning_file, "warning: ");
      vfprintf(warning_file, format, ap);
    }

  va_end(ap);
}

void error_print_debug(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  
  if (debug_file)
    {
      fprintf(debug_file, "debug: ");
      vfprintf(debug_file, format, ap);
    }

  va_end(ap);
}
