#ifndef ZASM_STREAM_H
#define ZASM_STREAM_H 1

#include <zasm/tokens.h>
#include <zasm/list.h>

#define STREAM_T(O) ((struct stream_t *)(O))
struct stream_t
{
  // public
  struct list_t list;
  // private
  struct action_t *cb_read;
  struct action_t *cb_wrote;
};

#define FEEDER_T(O) ((struct feeder_t *)(O))
struct feeder_t
{
  struct feeder_src_t *source;
  struct feeder_src_t *backup;
};

#define ACTION_T(O) ((struct action_t *)(O))
#define ACTION_EXECUTE_T(O) ((action_execute_t)(O))
typedef void (*action_execute_t)(struct feeder_t *input, struct stream_t *output, void *data);
struct action_t
{
  // private
  void *data;
  struct feeder_t *input;
  struct stream_t *output;
  action_execute_t execute;
};

#include <zasm/feeder.h>

struct action_t *action_new    (action_execute_t execute,
				struct feeder_t *input, struct stream_t *output, void *data);
void             action_fill   (struct action_t *action,
				action_execute_t execute,
				struct feeder_t *input, struct stream_t *output, void *data);
void             action_connect(struct action_t *action, struct feeder_t *input, struct stream_t *output);
void             action_execute(struct action_t *action);
void             action_delete (struct action_t *action);
void             action_free   (struct action_t *action);

struct stream_t *stream_new   (struct action_t *cb_read, struct action_t *cb_wrote);
void             stream_fill  (struct stream_t *stream, struct action_t *cb_read, struct action_t *cb_wrote);
struct token_t   stream_get   (struct stream_t *stream);
void             stream_unget (struct stream_t *stream, struct token_t token);
void             stream_put   (struct stream_t *stream, struct token_t token);
void             stream_delete(struct stream_t *stream);
void             stream_free  (struct stream_t *stream);

#endif
