#include <ti83plus.inc>
.export grayscale_show
.export grayscale_default_stop_init
.export grayscale_default_stop

;; grayscale_show
;;   description:	show a 4 level grayscale image
;;   comments:		interrupts are disabled during execution, but enabled when returned
;;   arguments:	ix:	the principal color map
;;		iy:	the secondary color map
;;		register hl:		a stop condition function
;;					  comments:	interrupts are disabled during execution, and it should take
;;							always exactely the same amount of time
;;                                        arguments:	none
;;					  result:	flag c:	stop
;;					  registers destroyed:	all except hl,ix,iy,b,d
;;   result:	none
;;   registers destroyed:	all,except ix,iy
grayscale_show:
.begin_section
	di
	ld d,0b10101010
	ld e,0b10101010
loop:
	push de
	call grayscale_show_planes	; ??
	pop de
	call call_hl			; ??
	jr c,stop			; 12/7
	rrc e
	rrc d
	push de
	call grayscale_show_planes	; ??
	pop de
	call call_hl			; ??
	jr c,stop			; 12/7
	rrc e
	jr loop				; 12
stop:
	ei
	ret
.end_section

;; grayscale_show_planes
;;   description:	show two planes interleaved in proportion 2:1
;;   arguments:		ix:	principal plane
;;			iy:	secondary plane
;;			d,e:	the current mask for the grayscales in the cycling
;;   result:	none
;;   registers destroyed:	a,f,c,b
grayscale_show_planes:
.begin_section
	;; set y-inc mode
	ld a,0x05		; 7	
	out (0x10),a
	ld c,0x20		; 7
	push ix			; 15
	push iy			; 15
	;; waste 15 ticks
	add a,0
	nop
	nop			; 4
	nop			; 4
	nop
x:
	;; set the x coordinate
	ld a,c			; 4
	out (0x10),a
	;; waste 53 ticks
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	add a,0			; 7
	nop			; 4	
	;; set the y coordinate to 0
	ld a,0x80		; 7
	out (0x10),a
	ld b,64			; 7
	;; waste 4 ticks
y:
	ld a,e			; 4
	and (iy+0)		; 19
	ld (make_byte+1),a	; 13
	ld a,e			; 4
	cpl			; 4
	or d			; 4
	and (ix+0)		; 19
make_byte:
	or 0x00			; 7
	out (0x11),a
	inc ix			; 10
	inc iy			; 10
	rrc e			; 8
	djnz y			; 13/8
	inc c			; 4
	ld a,c			; 4
	cp 0x2c			; 7
	jr nz,x			; 12/7
	pop iy
	pop ix
	ret
.end_section

;; grayscale_default_stop_init
;;   description: initialise the grayscale_default_stop function
;;   arguments:	none
;;   returns:	nothing
;;   registers destroyed:	a,f
grayscale_default_stop_init:
	ld a,0xff
	out (0x01),a
	ld a,0x00
	out (0x01),a
	in a,(0x01)
	ld (grayscale_default_stop_last_keys+1),a
	ret

;; grayscale_default_stop
;;   description: a default stop condition: is a key pressed?
;;   remarks:	
;;     to sift out the enter key that started the program, we ask that if a key was pressed the first time
;;     this function is called, it should be released first.
;;   arguments:	none
;;   returns:	flag c:	stop
;;   registers destroyed: a,f
grayscale_default_stop:
	ld a,0xff
	out (0x01),a
	ld a,0x00
	out (0x01),a
	in a,(0x01)
grayscale_default_stop_last_keys:
	cp 0x00
.begin_section
	jr z,no_stop1					; 12/7 
	ld (grayscale_default_stop_last_keys+1),a	; 13
	cp 0xff						; 7
	jr z,no_stop					; 12/7
	scf
	ret
no_stop1:
	;; waste 25 ticks
	push hl			; 10
	pop hl			; 11
	or a			; 4
no_stop:
	scf
	ccf
	ret
.end_section

call_hl:
	jp (hl)
