#include <ti83plus.inc>
#include "math.inc"
.org 9d95h
.begin_section
	push af
	push hl
	push de
	push bc
	push ix
	push iy
	
	;; parse str1
	ld hl,string1_name
	B_CALL Mov9ToOP1
	B_CALL FindSym
	jp c,not_found
	ld a,b
	or a
	jp nz,not_found
	ex de,hl

	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	call parser_parse_expr
	jp c,parse_error

	;; initialise the screen parameters
	;; var_x_min
	LOADREAL(2,Xmin)
	SAVEREAL(2,var_x_min)
	;; var_x_inc
	LOADREAL(1,Xmax)
	B_CALL FPSub
	ld a,95
	B_CALL SetXXOP2
	B_CALL FPDiv
	SAVEREAL(1,var_x_inc)
	call printOP1
	;; var_x_byte_inc
	ld a,8
	B_CALL SetXXOP2
	B_CALL FPMult
	SAVEREAL(1,var_x_byte_inc)
	call printOP1

	;; var_y_min
	LOADREAL(2,Ymin)
	SAVEREAL(2,var_y_min)
	;; var_y_inc
	LOADREAL(1,Ymax)
	B_CALL FPSub
	ld a,63
	B_CALL SetXXOP2
	B_CALL FPDiv
	SAVEREAL(1,var_y_inc)
	call printOP1

	;; make the image
	COPYREAL(var_x_min,var_x_byte)
	ld de,plane_1
	ld (location_plane_1),de
	ld de,plane_2
	ld (location_plane_2),de
	ld b,12
label_x:
	  push bc
	  COPYREAL(var_y_min,var_y)
	  ld a,'X'
	  B_CALL PutC
	  ld b,64
label_y:
	    push bc
	    COPYREAL(var_x_byte,var_x)
	    ld hl,var_y
	    ld de,parser_var_y
	    call ExportReal
	    ld b,8
label_byte:
	      push bc
	      ld hl,var_x
	      ld de,parser_var_x
	      call ExportReal
	      call parser_code
	      call add_bit
	      LOADREAL(1,var_x)
	      LOADREAL(2,var_x_inc)
	      B_CALL FPAdd
	      SAVEREAL(1,var_x)
	      pop bc
	      dec b
	      xor a
	      or b
	      jp nz,label_byte
	    ld hl,(location_plane_1)
	    inc hl
	    ld (location_plane_1),hl
	    ld hl,(location_plane_2)
	    inc hl
	    ld (location_plane_2),hl
	    LOADREAL(1,var_y)
	    LOADREAL(2,var_y_inc)
	    B_CALL FPAdd
	    SAVEREAL(1,var_y)
	    pop bc
	    dec b
	    xor a
	    or b
	    jp nz,label_y
	  COPYREAL(var_x,var_x_byte)
	  pop bc
	  dec b
	  xor a
	  or b
	  jp nz,label_x

	;; and, finally, show
	di
	push ix
	push iy
	call grayscale_default_stop_init
	ld hl,grayscale_default_stop
	ld ix,plane_1
	ld iy,plane_2
	call grayscale_show
	di
	pop iy
	pop ix
	ei

	B_CALL ClrLCDFull

end:
	pop iy
	pop ix
	pop bc
	pop de
	pop hl
	pop af
	ret

not_found:
	ld hl,err_string1_not_found
	B_CALL PutS
	jr end

parse_error:
	ld hl,err_parse_error
	B_CALL PutS
	jr end
.end_section

add_bit:
.begin_section
	ld hl,OP1+1
	ld a,(hl)
	inc hl
	cp 0x7f+14
	jr nc,overflow
	sub 0x7f
	jr c,almost_0
	rra
	jr c,odd
	ld b,0
	ld c,a
	add hl,bc
	ld a,(hl)
	jr do_bit
odd:
	ld b,0
	ld c,a
	add hl,bc
	ld a,(hl)
	inc hl
	ld b,(hl)
	rl b
	rla
	rl b
	rla
	rl b
	rla
	rl b
	rla
	jr do_bit
overflow:
	ld a,0x99
	jr do_bit
almost_0:
	ld a,0x00
do_bit:
	ld hl,OP1
	bit 7,(hl)
	jr z,positive
	neg
	daa
positive:
	ld hl,(location_plane_1)
	add a,a
	daa
	rl (hl)
	ld hl,(location_plane_2)
	add a,a
	daa
	rl (hl)
	ret
.end_section

printOP1:
.begin_section
	B_CALL PushRealO1
	B_CALL PushRealO2
	B_CALL PushRealO3
	B_CALL PushRealO4
	ld a,15
	B_CALL FormReal
	ld hl,OP3
	B_CALL PutS
	ld a,(curRow)
	inc a
	and 0x07
	ld (curRow),a
	ld a,0
	ld (curCol),a
	B_CALL PopRealO4
	B_CALL PopRealO3
	B_CALL PopRealO4
	B_CALL PopRealO1
	ret
.end_section

err_string1_not_found:	.db "error: string 1 not found", 0
err_parse_error:	.db "error: string 1 not found", 0

string1_name:
	.db StrngObj, tVarStrng, tStr1, 0, 0, 0, 0, 0, 0

var_x:		.db [9]0x00
var_x_min:	.db [9]0x00
var_x_inc:	.db [9]0x00
var_x_byte:	.db [9]0x00
var_x_byte_inc:	.db [9]0x00
var_y:		.db [9]0x00
var_y_inc:	.db [9]0x00
var_y_min:	.db [9]0x00

location_plane_1:
	.dw plane_1
location_plane_2:
	.dw plane_2

const05:
	.db 0x00, 0x7F, 0x50, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	
plane_1:
	.db [768]0x00
plane_2:
	.db [768]0x00